import json
import re
import click
from   .api import QyPIError

def parse_packages(ctx, packages, pre=False, newest=False, versioned=True):
    ### TODO: Figure out a better way to integrate this with Click
    ok = True
    for pkgname in packages:
        try:
            if versioned:
                name, eq, version = pkgname.partition('=')
                if eq == '':
                    pkg = ctx.obj.get_latest_version(name, pre, newest)
                else:
                    pkg = ctx.obj.get_version(name, version.lstrip('='))
            else:
                pkg = ctx.obj.get_package(pkgname)
        except QyPIError as e:
            click.echo(ctx.command_path + ': ' + str(e), err=True)
            ok = False
        else:
            yield pkg
    if not ok:
        ctx.exit(1)

def dumps(obj):
    return json.dumps(obj, sort_keys=True, indent=4, ensure_ascii=False)

def clean_pypi_dict(d):
    return {
        k: (None if v in ('', 'UNKNOWN') else v)
        for k,v in d.items() if not k.startswith(('cheesecake', '_pypi'))
    }


class JSONLister:
    def __init__(self):
        self.first = True

    def __enter__(self):
        click.echo('[', nl=False)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.first:
            click.echo()
        click.echo(']')
        return False

    def append(self, obj):
        if self.first:
            click.echo()
            self.first = False
        else:
            click.echo(',')
        click.echo(re.sub(r'^', '    ', dumps(obj), flags=re.M), nl=False)


class JSONMapper:
    def __init__(self):
        self.first = True

    def __enter__(self):
        click.echo('{', nl=False)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.first:
            click.echo()
        click.echo('}')
        return False

    def append(self, key, value):
        if self.first:
            click.echo()
            self.first = False
        else:
            click.echo(',')
        click.echo(
            re.sub(r'^', '    ', json.dumps(key)+': '+dumps(value), flags=re.M),
            nl=False,
        )



from   xmlrpc.client     import ServerProxy
from   packaging.version import parse
import requests

class PyPIClient:
    def __init__(self, index_url):
        self.index_url = index_url
        self.s = None
        self.xsp = None

    def get(self, *path):
        if self.s is None:
            self.s = requests.Session()
        return self.s.get(self.index_url.rstrip('/') + '/' + '/'.join(path))

    def get_package(self, package):
        r = self.get(package, 'json')
        # Unlike the XML-RPC API, the JSON API accepts package names regardless
        # of normalization
        if r.status_code == 404:
            raise PackageNotFoundError(package)
        r.raise_for_status()
        return r.json()

    def get_latest_version(self, package, pre=False, newest=False):
        pkg = self.get_package(package)
        releases = {
            parse(rel): first_upload(files)
            for rel, files in pkg["releases"].items()
        }
        candidates = releases.keys()
        if not pre:
            candidates = filter(lambda v: not v.is_prerelease, candidates)
        if newest:
            latest = max(
                filter(releases.__getitem__, candidates),
                key=releases.__getitem__,
                default=None,
            )
        else:
            latest = max(candidates, default=None)
        if latest is None:
            raise NoSuitableVersionError(package)
        latest = str(latest)
        ### TODO: Will stringifying the parsed version string instead of using
        ### the original key from `pkg["releases"]` ever change the version
        ### string in a meaningful way?
        if pkg["info"]["version"] == latest:
            return pkg
        else:
            return self.get_version(package, latest)

    def get_version(self, package, version):
        r = self.get(package, version, 'json')
        if r.status_code == 404:
            raise VersionNotFoundError(package, version)
        r.raise_for_status()
        return r.json()

    def xmlrpc(self, method, *args, **kwargs):
        if self.xsp is None:
            self.xsp = ServerProxy(self.index_url)
        return getattr(self.xsp, method)(*args, **kwargs)


class QyPIError(Exception):
    pass


class PackageNotFoundError(QyPIError):
    def __init__(self, package):
        self.package = package

    def __str__(self):
        return self.package + ': package not found'


class VersionNotFoundError(QyPIError):
    def __init__(self, package, version):
        self.package = package
        self.version = version

    def __str__(self):
        return '{0.package}: version {0.version} not found'.format(self)


class NoSuitableVersionError(QyPIError):
    def __init__(self, package):
        self.package = package

    def __str__(self):
        return self.package + ': no suitable versions available'

def first_upload(files):
    return min((f["upload_time"] for f in files), default=None)



from setuptools import setup, find_packages

setup(
    name='qypi',
    #version - in setup.cfg
    packages=find_packages(),
    license='MIT',
    author='John Thorvald Wodder II',
    author_email='qypi@varonathe.org',
    keywords='pypi warehouse search packages pip',
    description='Query PyPI from the command line',
    #long_description - in setup.cfg
    url='https://github.com/jwodder/qypi',

    python_requires='~=3.4',
    install_requires=['click~=6.5', 'packaging>=16', 'requests==2.*'],

    classifiers=[
        'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Software Distribution',
    ],

    entry_points={
        "console_scripts": [
            'qypi = qypi.__main__:qypi',
        ]
    },
)
